"""Parse DFD source text in our DSL"""

import os.path
import re
from typing import Tuple

from . import dfd_dot_templates as TMPL
from . import model
from .console import dprint


def check(statements: model.Statements) -> dict[str, model.Item]:
    """Check that all statement make sense"""

    # collect items
    items_by_name: dict[str, model.Item] = {}
    for statement in statements:
        error_prefix = model.mk_err_prefix_from(statement.source)
        match statement:
            case model.Item() as item:
                pass
            case _:
                continue
        name = item.name

        # make sure there are no duplicates
        if name not in items_by_name:
            items_by_name[name] = item
            continue

        other = items_by_name[name]
        other_text = model.pack(other.source.text)
        raise model.DfdException(
            f'{error_prefix}Name "{name}" already exists '
            f"at line {other.source.line_nr+1}: {other_text}"
        )

    # check references and values of connections
    for statement in statements:
        error_prefix = model.mk_err_prefix_from(statement.source)
        match statement:
            case model.Connection() as conn:
                pass
            case _:
                continue
        nb_stars = 0
        for point in conn.src, conn.dst:
            if point == "*":
                nb_stars += 1
            if point != "*":
                if point not in items_by_name:
                    raise model.DfdException(
                        f'{error_prefix}Connection "{conn.type}" links to "{point}", '
                        f"which is not defined"
                    )
                if (
                    items_by_name[point].type == model.CONTROL
                    and conn.type != model.SIGNAL
                ):
                    raise model.DfdException(
                        f'{error_prefix}Connection to {model.CONTROL} "{point}" is '
                        f'of type "{conn.type}", however only connections of type '
                        f'"{model.SIGNAL}" are allowed'
                    )

        if nb_stars == 2:
            raise model.DfdException(
                f'{error_prefix}Connection "{conn.type}" may not link to two '
                f"stars"
            )

    # check references of frames
    framed_items: set[str] = set()
    for statement in statements:
        error_prefix = model.mk_err_prefix_from(statement.source)
        match statement:
            case model.Frame() as frame:
                pass
            case _:
                continue
        if not frame.items:
            raise model.DfdException(f"{error_prefix}Frame is empty")
        for name in frame.items:
            if name not in items_by_name:
                raise model.DfdException(
                    f'{error_prefix}Frame includes "{name}", '
                    f"which is not defined"
                )
            if name in framed_items:
                raise model.DfdException(
                    f'{error_prefix}Item "{name}", ' f"is in multiple frames"
                )
            framed_items.add(name)

    return items_by_name


def parse(
    source_lines: model.SourceLines,
    debug: bool = False,
) -> tuple[model.Statements, model.GraphDependencies, model.Attribs]:
    """Parse the DFD source text as list of statements"""

    statements: model.Statements = []
    dependencies: model.GraphDependencies = []
    attribs: model.Attribs = {}

    for n, source in enumerate(source_lines):
        src_line = source.text

        src_line = src_line.strip()
        if not src_line or src_line.startswith("#"):
            continue
        error_prefix = model.mk_err_prefix_from(source)

        # syntactic sugars may rewrite the line
        line = apply_syntactic_sugars(src_line)
        source.text = line  # fixup text

        word = source.text.split()[0]

        f = {
            model.STYLE: parse_style,
            model.PROCESS: parse_process,
            model.CONTROL: parse_control,
            model.ENTITY: parse_entity,
            model.STORE: parse_store,
            model.NONE: parse_none,
            model.CHANNEL: parse_channel,
            model.FLOW: parse_flow,
            model.CFLOW: parse_cflow,
            model.BFLOW: parse_bflow,
            model.UFLOW: parse_uflow,
            model.SIGNAL: parse_signal,
            model.FLOW + "?": parse_flow_q,
            model.CFLOW + "?": parse_cflow_q,
            model.BFLOW + "?": parse_bflow_q,
            model.UFLOW + "?": parse_uflow_q,
            model.SIGNAL + "?": parse_signal_q,
            "flow.r": parse_flow_r,
            "cflow.r": parse_cflow_r,
            "signal.r": parse_signal_r,
            "flow.r?": parse_flow_r_q,
            "cflow.r?": parse_cflow_r_q,
            "signal.r?": parse_signal_r_q,
            model.FRAME: parse_frame,
            model.ATTRIB: parse_attrib,
            model.ONLY: parse_filter,
            model.WITHOUT: parse_filter,
        }.get(word)

        if f is None:
            raise model.DfdException(
                f"{error_prefix}Unrecognized keyword " f'"{word}"'
            )

        try:
            statement = f(source)
        except model.DfdException as e:
            raise model.DfdException(f'{error_prefix}{e}')

        match statement:
            case model.Item() as item:
                parse_item_external(item, dependencies)
                item.text = item.text or item.name

        match statement:
            case model.Drawable() as drawable:
                parse_drawable_attrs(drawable)

        match statement:
            case model.Attrib() as attrib:
                attribs[attrib.alias] = attrib

        statements.append(statement)

    if debug:
        for s in statements:
            dprint(model.repr(s))
    return statements, dependencies, attribs


def split_args(
    dfd_line: str, n: int, last_is_optional: bool = False
) -> list[str]:
    """Split DFD line into n (possibly n-1) tokens"""

    terms: list[str] = dfd_line.split(maxsplit=n)
    if len(terms) - 1 == n - 1 and last_is_optional:
        terms.append("")

    if len(terms) - 1 != n:
        if not last_is_optional:
            raise model.DfdException(f"Expected {n} argument(s)")
        else:
            raise model.DfdException(f"Expected {n-1} or {n} argument")

    return terms[1:]


def parse_item_name(name: str) -> Tuple[str, bool]:
    """If name ends with ?, make it hidable"""
    if name.endswith("?"):
        return name[:-1], True
    else:
        return name, False


def parse_style(source: model.SourceLine) -> model.Statement:
    """Parse style statement"""
    style, value = split_args(source.text, 2, True)
    return model.Style(source, style, value)


def parse_attrib(source: model.SourceLine) -> model.Statement:
    """Parse attrib name text"""
    alias, text = split_args(source.text, 2, True)
    return model.Attrib(source, alias, text)


RX_FILTER_ARG = re.compile(
    r"""(
      # either a neighbor specification
      (?P<neighbors><>|<|>|\[|])    # direction
      (?P<flags>[a-zA-Z]*)              # flags
      (?: (?P<all>[*]) | (?P<num>[0-9]+) )  # "*" for all, or decimal number
      |
      # or a replacer specification
      =                             # indicates replacer
      (?P<replacer>.*)              # name if item replacing the others
    )""",
    re.X,
)


def parse_filter(source: model.SourceLine) -> model.Statement:
    """Parse !/~[NEIGHBOURS] NAME[S]"""
    terms: list[str] = source.text.split()
    if len(terms) < 2:
        raise model.DfdException(f"One or more arguments are expected")

    f = model.Filter(
        source,
        names=[],
        neighbors_up=model.FilterNeighbors(0, False, False, False),
        neighbors_down=model.FilterNeighbors(0, False, False, False),
    )

    cmd = terms[0]
    args = terms[1:]
    replacer = ""

    # first arguments may be +N or -N, which specify the number of down/up neighbors to include in the filter
    while args:
        arg = args[0]

        match = RX_FILTER_ARG.fullmatch(arg)
        if not match:
            break  # no neighbor/replacer specification

        if match.group("replacer"):
            if cmd != model.WITHOUT:
                raise model.DfdException(
                    f"Replacer specification is only allowed for {model.WITHOUT} filter"
                )
            replacer = arg[1:]
            args = args[1:]
        elif match.group("neighbors"):
            fn = model.FilterNeighbors(0, False, False, False)
            is_up = is_down = False

            match match.group("neighbors"):
                case "<>":
                    is_up = is_down = True
                case "<":
                    is_up = True
                case ">":
                    is_down = True
                case "[":
                    is_up = True
                    fn.layout_dir = True
                case "]":
                    is_down = True
                    fn.layout_dir = True

            for flag in match.group("flags"):
                match flag:
                    case "x":
                        fn.no_anchors = True
                    case "f":
                        fn.no_frames = True
                    case _:
                        raise model.DfdException(
                            f"Unrecognized filter flag: {flag}"
                        )
            if match.group("all"):
                fn.distance = -1  # special value for "all neighbors"
            elif match.group("num"):
                fn.distance = int(match.group("num"))
            else:
                raise model.DfdException(
                    f"Neighborhood size must be an integer or '*', not: {arg}"
                )

            if is_up:
                f.neighbors_up = fn
            if is_down:
                f.neighbors_down = fn

            # ready for next argument
            args = args[1:]

    if len(args) == 0:
        raise model.DfdException(f"One or more names are expected")

    f.names = args

    res: model.Statement
    if cmd == model.ONLY:
        res = model.Only(**f.__dict__)
    else:  # cmd == model.WITHOUT:
        res = model.Without(
            **f.__dict__, replaced_by=replacer
        )  # replaced_by is set later by the caller
    return res


def parse_process(source: model.SourceLine) -> model.Statement:
    """Parse process statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.PROCESS, text, "", name, hidable)


def parse_control(source: model.SourceLine) -> model.Statement:
    """Parse control statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.CONTROL, text, "", name, hidable)


def parse_entity(source: model.SourceLine) -> model.Statement:
    """Parse entity statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.ENTITY, text, "", name, hidable)


def parse_store(source: model.SourceLine) -> model.Statement:
    """Parse store statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.STORE, text, "", name, hidable)


def parse_none(source: model.SourceLine) -> model.Statement:
    """Parse none statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.NONE, text, "", name, hidable)


def parse_channel(source: model.SourceLine) -> model.Statement:
    """Parse channel statement"""
    name, text = split_args(source.text, 2, True)
    name, hidable = parse_item_name(name)
    return model.Item(source, model.CHANNEL, text, "", name, hidable)


def parse_flow(source: model.SourceLine) -> model.Statement:
    """Parse directional flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.FLOW, text, "", src, dst)


def parse_flow_r(source: model.SourceLine) -> model.Statement:
    """Parse directional reversed flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.FLOW, text, "", src, dst, True)


def parse_cflow(source: model.SourceLine) -> model.Statement:
    """Parse continuous flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.CFLOW, text, "", src, dst)


def parse_cflow_r(source: model.SourceLine) -> model.Statement:
    """Parse continuous flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.CFLOW, text, "", src, dst, True)


def parse_bflow(source: model.SourceLine) -> model.Statement:
    """Parse bidirectional flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.BFLOW, text, "", src, dst)


def parse_uflow(source: model.SourceLine) -> model.Statement:
    """Parse undirected flow flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.UFLOW, text, "", src, dst)


def parse_signal(source: model.SourceLine) -> model.Statement:
    """Parse signal statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.SIGNAL, text, "", src, dst)


def parse_signal_r(source: model.SourceLine) -> model.Statement:
    """Parse reversed signal statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(source, model.SIGNAL, text, "", src, dst, True)


def parse_flow_q(source: model.SourceLine) -> model.Statement:
    """Parse directional flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(
        source, model.FLOW, text, "", src, dst, relaxed=True
    )


def parse_flow_r_q(source: model.SourceLine) -> model.Statement:
    """Parse directional reversed flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(
        source, model.FLOW, text, "", src, dst, True, relaxed=True
    )


def parse_cflow_q(source: model.SourceLine) -> model.Statement:
    """Parse continuous flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(
        source, model.CFLOW, text, "", src, dst, relaxed=True
    )


def parse_cflow_r_q(source: model.SourceLine) -> model.Statement:
    """Parse continuous flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(
        source, model.CFLOW, text, "", src, dst, True, relaxed=True
    )


def parse_bflow_q(source: model.SourceLine) -> model.Statement:
    """Parse bidirectional flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(
        source, model.BFLOW, text, "", src, dst, relaxed=True
    )


def parse_uflow_q(source: model.SourceLine) -> model.Statement:
    """Parse undirected flow flow statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(
        source, model.UFLOW, text, "", src, dst, relaxed=True
    )


def parse_signal_q(source: model.SourceLine) -> model.Statement:
    """Parse signal statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(
        source, model.SIGNAL, text, "", src, dst, relaxed=True
    )


def parse_signal_r_q(source: model.SourceLine) -> model.Statement:
    """Parse reversed signal statement"""
    src, dst, text = split_args(source.text, 3, True)
    return model.Connection(
        source, model.SIGNAL, text, "", src, dst, True, relaxed=True
    )


def apply_syntactic_sugars(src_line: str) -> str:

    # allow arguments to be sticked to the filter mnemonics
    if src_line and src_line[0] in ("!~"):
        if len(src_line) > 1 and src_line[1] != " ":
            # insert a space after the filter, so that it is recognized as a filter
            new_line = src_line[0] + " " + src_line[1:]
            return new_line

    # others syntactic sugars apply on 3 or more terms
    terms = src_line.split()
    if len(terms) < 3:
        return src_line

    op = terms[1]

    def fmt(verb: str, args: list[str], swap: bool = False) -> str:
        array = [verb] + args
        del array[2]  # remove arrow
        if swap:
            array[1], array[2] = array[2], array[1]
        return "\t".join(array)

    new_line = ""
    if re.fullmatch(r"-+>[?]?", op):
        q = "?" if op.endswith("?") else ""
        parts = src_line.split(maxsplit=3)
        new_line = fmt("flow" + q, parts)
    elif re.fullmatch(r"<-+[?]?", op):
        q = "?" if op.endswith("?") else ""
        parts = src_line.split(maxsplit=3)
        new_line = fmt("flow.r" + q, parts)  # , swap=True)

    if re.fullmatch(r"-+>>[?]?", op):
        q = "?" if op.endswith("?") else ""
        parts = src_line.split(maxsplit=3)
        new_line = fmt("cflow" + q, parts)
    elif re.fullmatch(r"<<-+[?]?", op):
        q = "?" if op.endswith("?") else ""
        parts = src_line.split(maxsplit=3)
        new_line = fmt("cflow.r" + q, parts)  # , swap=True)

    elif re.fullmatch(r"<-+>[?]?", op):
        q = "?" if op.endswith("?") else ""
        parts = src_line.split(maxsplit=3)
        new_line = fmt("bflow" + q, parts)

    elif re.fullmatch(r"--+[?]?", op):
        q = "?" if op.endswith("?") else ""
        parts = src_line.split(maxsplit=3)
        new_line = fmt("uflow" + q, parts)

    elif re.fullmatch(r":+>[?]?", op):
        q = "?" if op.endswith("?") else ""
        parts = src_line.split(maxsplit=3)
        new_line = fmt("signal" + q, parts)
    elif re.fullmatch(r"<:+[?]?", op):
        q = "?" if op.endswith("?") else ""
        parts = src_line.split(maxsplit=3)
        new_line = fmt("signal.r" + q, parts)  # , swap=True)

    if new_line:
        return new_line
    else:
        return src_line


def parse_drawable_attrs(drawable: model.Drawable) -> None:
    if drawable.text and drawable.text.startswith("["):
        parts = drawable.text[1:].split("]", 1)
        drawable.attrs = parts[0]
        drawable.text = parts[1].strip()

        match drawable:
            case model.Item() as item:
                item.text = item.text or item.name


def parse_item_external(
    item: model.Item, dependencies: model.GraphDependencies
) -> None:
    parts = item.name.split(":", 1)
    if len(parts) > 1:
        item.attrs = TMPL.ITEM_EXTERNAL_ATTRS
        if parts[-1]:
            item.name = parts[-1]
        else:
            item.name = parts[-2]

        if item.name.startswith("#"):
            item.name = item.name[1:]
        else:
            item.name = os.path.splitext(item.name)[0]

        if not item.text:
            item.text = item.name

        dependency = model.GraphDependency(
            parts[0], parts[1] or None, item.type, item.source
        )
        dependencies.append(dependency)


def parse_frame(source: model.SourceLine) -> model.Statement:
    """Parse frame statement"""
    parts = source.text.split("=", maxsplit=1)
    if len(parts) == 1:
        text = ""
    else:
        text = parts[1].strip()

    items = parts[0].split()[1:]
    type = ""  # so far there is only one type of frame
    attrs = "style=dashed"
    return model.Frame(source, type, text, attrs, items)



import sys
from typing_extensions import Literal


def print_error(text: str) -> None:
    if sys.stderr.isatty():
        text = f"\033[31m{text}\033[0m"
    print(text, file=sys.stderr)


debug = False


def set_debug(value: bool) -> None:
    global debug
    debug = value


def dprint(
    *values: object,
    sep: str | None = " ",
    end: str | None = "\n",
    flush: Literal[False] = False,
) -> None:
    if not debug:
        return
    print(*values, sep=sep, end=end, flush=flush, file=sys.stderr)



"""Command-line DFD diagram generator. Converts a textual description
into a graphic file.

-----

See https://github.com/pbauermeister/dfd for information, syntax and
examples.

-----

This module parses the commandline args, prepares the I/Os and calls the stuff.
"""

import argparse
import os
import sys
import tempfile
from typing import TextIO

import pkg_resources

from . import dfd, dot, markdown, model
from .console import print_error, dprint, set_debug

try:
    VERSION = pkg_resources.require("data-flow-diagram")[0].version
except pkg_resources.DistributionNotFound:
    VERSION = "undefined"


def parse_args() -> argparse.Namespace:
    assert __doc__ is not None
    description, epilog = [each.strip() for each in __doc__.split("-----")[:2]]

    parser = argparse.ArgumentParser(description=description, epilog=epilog)

    parser.add_argument(
        "INPUT_FILE",
        action="store",
        default=None,
        nargs="?",
        help="UML sequence input file; " "if omitted, stdin is used",
    )

    parser.add_argument(
        "--output-file",
        "-o",
        required=False,
        help="output file name; pass '-' to use stdout; "
        "if omitted, use INPUT_FILE base name with '.svg' "
        "extension, or stdout",
    )

    parser.add_argument(
        "--markdown",
        "-m",
        action="store_true",
        help="consider snippets between opening marker: "
        "```data-flow-diagram OUTFILE, and closing marker: ``` "
        "allowing to generate all diagrams contained in an "
        "INPUT_FILE that is a markdown file",
    )

    parser.add_argument(
        "--format",
        "-f",
        required=False,
        default="svg",
        help="output format: gif, jpg, tiff, bmp, pnm, eps, "
        "pdf, svg (any supported by Graphviz); default is svg",
    )

    parser.add_argument(
        "--percent-zoom",
        "-p",
        required=False,
        default=100,
        type=int,
        help="magnification percentage; default is 100",
    )

    parser.add_argument(
        "--background-color",
        "-b",
        required=False,
        default="white",
        help="background color name (including 'none' for"
        " transparent) in web color notation; see"
        " https://developer.mozilla.org/en-US/docs/Web/CSS/color_value"
        " for a list of valid names; default is white",
    )

    parser.add_argument(
        "--no-graph-title",
        action="store_true",
        default=False,
        help="suppress graph title",
    )

    parser.add_argument(
        "--no-check-dependencies",
        action="store_true",
        default=False,
        help="suppress dependencies checking",
    )

    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="emit debug messages",
    )

    parser.add_argument(
        "--version",
        "-V",
        action="store_true",
        help="print the version and exit",
    )

    return parser.parse_args()


def handle_markdown_source(
    options: model.Options, provenance: str, input_fp: TextIO
) -> None:
    text = input_fp.read()
    snippets = markdown.extract_snippets(text)
    markdown.check_snippets_unicity(provenance, snippets)
    snippets_params = markdown.make_snippets_params(provenance, snippets)
    for params in snippets_params:
        dfd.build(
            params.root,
            params.input_fp.read(),
            params.file_name,
            options,
            snippet_by_name=params.snippet_by_name,
        )
        dprint(f"{sys.argv[0]}: generated {params.file_name}")


def handle_dfd_source(
    options: model.Options, provenance: str, input_fp: TextIO, output_path: str
) -> None:
    root = model.SourceLine("", provenance, None, 0)
    if output_path == "-":
        # output to stdout
        with tempfile.TemporaryDirectory() as d:
            path = os.path.join(d, "file.svg")
            dfd.build(root, input_fp.read(), path, options)
            with open(path) as f:
                print(f.read())
    else:
        # output to file
        dfd.build(root, input_fp.read(), output_path, options)


def run(args: argparse.Namespace) -> None:
    # adjust input
    if args.INPUT_FILE is None:
        input_fp = sys.stdin
        provenance = "<stdin>"
    else:
        input_fp = open(args.INPUT_FILE)
        provenance = f"<file:{args.INPUT_FILE}>"

    options = model.Options(
        args.format,
        args.percent_zoom,
        args.background_color,
        args.no_graph_title,
        args.no_check_dependencies,
        args.debug,
    )

    set_debug(args.debug)

    # markdown source
    if args.markdown:
        handle_markdown_source(options, provenance, input_fp)
        return

    # adjust output
    if args.output_file is None:
        if args.INPUT_FILE is not None:
            basename = os.path.splitext(args.INPUT_FILE)[0]
            output_path = basename + "." + args.format
        else:
            output_path = "-"
    else:
        output_path = args.output_file

    # DFD source
    handle_dfd_source(options, provenance, input_fp, output_path)


def main() -> None:
    """Entry point for the application script"""

    dot.check_installed()

    args = parse_args()
    if args.version:
        print("data-flow-diagram", VERSION)
        sys.exit(0)

    try:
        run(args)
    except model.DfdException as e:
        text = f"ERROR: {e}"
        print_error(text)
        sys.exit(1)



"""Internal data model"""

from __future__ import annotations

import dataclasses
import json
from dataclasses import dataclass
from typing import Any

from . import config


def repr(o: Any) -> str:
    name: str = o.__class__.__name__
    val: str = json.dumps(dataclasses.asdict(o), indent="  ")
    return f"{name} {val}"


@dataclass
class Base:
    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + " "
            + json.dumps(dataclasses.asdict(self), indent="  ")
        )


@dataclass
class Snippet(Base):
    text: str
    name: str
    output: str
    line_nr: int


@dataclass
class SourceLine(Base):
    text: str  # after pre-processor
    raw_text: str | None
    parent: SourceLine | None  # https://stackoverflow.com/a/62521947
    line_nr: int
    is_container: bool = False


@dataclass
class Statement(Base):
    source: SourceLine


@dataclass
class Drawable(Statement):
    type: str
    text: str
    attrs: str


@dataclass
class Item(Drawable):
    name: str
    hidable: bool


@dataclass
class Connection(Drawable):
    src: str
    dst: str
    reversed: bool = False
    relaxed: bool = False

    def signature(self) -> str:
        d = dataclasses.asdict(self).copy()
        del d["source"]
        return json.dumps(d, sort_keys=True)


@dataclass
class Frame(Drawable):
    items: list[str]


@dataclass
class Style(Statement):
    style: str
    value: Any = None


@dataclass
class Attrib(Statement):
    alias: str
    text: str


@dataclass
class FilterNeighbors:
    distance: int  # neighborhood distance to keep, or -1 for all
    no_anchors: bool  # consider only neighbors, not listed nodes themselves
    layout_dir: bool  # consider layout direction, not flow direction
    no_frames: bool  # suppress impacted frames


@dataclass
class Filter(Statement):
    names: list[str]
    neighbors_up: FilterNeighbors
    neighbors_down: FilterNeighbors


@dataclass
class Only(Filter):
    pass


@dataclass
class Without(Filter):
    replaced_by: str


Attribs = dict[str, Attrib]


STYLE = "style"

PROCESS = "process"
CONTROL = "control"
ENTITY = "entity"
STORE = "store"
CHANNEL = "channel"
NONE = "none"

FLOW = "flow"
BFLOW = "bflow"
CFLOW = "cflow"
UFLOW = "uflow"
SIGNAL = "signal"

FRAME = "frame"

ATTRIB = "attrib"

ONLY = "!"
WITHOUT = "~"


@dataclass
class GraphOptions:
    is_vertical: bool = False
    is_context: bool = False
    is_rotated: bool = False
    item_text_width = config.DEFAULT_ITEM_TEXT_WIDTH
    connection_text_width = config.DEFAULT_CONNECTION_TEXT_WIDTH


def pack(src_line: str | None) -> str:
    if src_line is None:
        return "<none>"
    return " ".join(src_line.split())


def mk_err_prefix_from(src: SourceLine) -> str:

    def _add_to_stack(stack: list[str], src: SourceLine) -> None:
        if src.line_nr is None:
            stack += [f"  {pack(src.raw_text)}"]
        else:
            if src.parent and src.parent.is_container:
                nr = src.parent.line_nr + 1
                delta = src.line_nr + 1
                final = nr + delta
                stack += [f"  line {final}: {pack(src.raw_text)}"]
            else:
                nr = src.line_nr + 1
                stack += [f"  line {nr}: {pack(src.raw_text)}"]
        if src.parent:
            _add_to_stack(stack, src.parent)

    stack: list[str] = ["(most recent first)"]
    _add_to_stack(stack, src)
    stack += [""]
    return "\n".join(stack) + "Error: "


class DfdException(Exception):
    pass


# Handy type aliases
Snippets = list[Snippet]
SourceLines = list[SourceLine]
Statements = list[Statement]
SnippetByName = dict[str, Snippet]


@dataclass
class Options:
    format: str
    percent_zoom: int
    background_color: str
    no_graph_title: bool
    no_check_dependencies: bool
    debug: bool


@dataclass
class GraphDependency:
    to_graph: str
    to_item: str | None
    to_type: str
    source: SourceLine


GraphDependencies = list[GraphDependency]



import io
import os
import re
from collections import Counter
from dataclasses import dataclass
from typing import TextIO

from . import model


@dataclass
class SnippetContext:
    root: model.SourceLine
    input_fp: TextIO
    file_name: str
    snippet_by_name: model.SnippetByName


SnippetContexts = list[SnippetContext]


def extract_snippets(text: str) -> model.Snippets:
    rx = re.compile(
        r"^```(?P<head>\s*)"
        r"data-flow-diagram\s+"
        r"(?P<output>.*?)\s*"
        r"^(?P<src>.*?)^\s*```",
        re.DOTALL | re.M,
    )

    return [
        model.Snippet(
            text=match["head"] + match["src"],
            name=os.path.splitext(match["output"])[0],
            output=match["output"],
            line_nr=len(text[: match.start()].splitlines()),
        )
        for match in rx.finditer(text)
    ]
    # FIXME: have text a TextIO and find in streaming


def make_snippets_params(
    provenance: str,
    snippets: model.Snippets,
) -> SnippetContexts:
    res: SnippetContexts = []
    snippet_by_name = {s.name: s for s in snippets}

    for snippet in snippets:
        # snippet w/o output, maybe just as includee
        if snippet.output.startswith("#"):
            continue

        # snippet with output
        input_fp = io.StringIO(snippet.text)
        snippet_provenance = f"{provenance}<snippet:{snippet.output}>"
        root = model.SourceLine(
            "", snippet_provenance, None, snippet.line_nr, is_container=True
        )
        file_name = input_fp.name = snippet.output
        res.append(SnippetContext(root, input_fp, file_name, snippet_by_name))

    return res


def check_snippets_unicity(provenance: str, snippets: model.Snippets) -> None:
    snippet_names = [s.name for s in snippets]  # capture duplicates
    counts = Counter(snippet_names)
    multiples = {k: n for k, n in counts.items() if n > 1}
    if multiples:
        root = model.SourceLine("", provenance, None, None)
        error_prefix = model.mk_err_prefix_from(root)
        raise model.DfdException(
            f"{error_prefix}Snippets defined multiple " f"times: {multiples}"
        )



STORE = """
"{name}" [shape=none label=<
  <TABLE BORDER="0">
    <TR><TD BGCOLOR="{color}" WIDTH="6"></TD></TR>
    <TR><TD><FONT COLOR="{fontcolor}">{text}</FONT></TD></TR>
    <TR><TD BGCOLOR="{color}" WIDTH="6"></TD></TR>
  </TABLE>>]
""".strip()


CHANNEL = """
"{name}" [shape=none label=<
  <TABLE BORDER="0">
    <TR>
      <TD WIDTH="48"></TD>
      <TD BGCOLOR="{color}" WIDTH="0" PORT="x"><BR/><BR/></TD>
      <TD WIDTH="48"></TD>
    </TR>
    <TR>
      <TD COLSPAN="3"><FONT COLOR="{fontcolor}">{text}</FONT></TD>
    </TR>
  </TABLE>>]
""".strip()


CHANNEL_HORIZONTAL = """
"{name}" [shape=none label=<
  <TABLE BORDER="0">
    <TR><TD BGCOLOR="{color}" PORT="x"></TD></TR>
    <TR><TD><FONT COLOR="{fontcolor}">{text}</FONT></TD></TR>
  </TABLE>>]
""".strip()


DOT_FONT_EDGE = 'fontname="times-italic" fontsize=10'
DOT_FONT_NODE = 'fontname="helvetica" fontsize=10'
DOT_FONT_GRAPH = 'fontname="helvetica" fontsize=9 fontcolor="#000060"'

DOT_GRAPH_TITLE = """graph[label="\n- {title} -" """ + DOT_FONT_GRAPH + """]"""
DOT_GRAPH_NOTITLE = f"graph[{DOT_FONT_GRAPH}]"

DOT = (
    """
digraph D {{
  {graph_params}
  edge[color=gray """
    + DOT_FONT_EDGE
    + """]
  node["""
    + DOT_FONT_NODE
    + """]
  {block}
}}
""".strip()
)


GRAPH_PARAMS_CONTEXT_DIAGRAM = "edge [len=2.25]"


ITEM_EXTERNAL_ATTRS = "fillcolor=white color=grey fontcolor=grey"



import os
import re

from . import model
from .console import dprint

# Regex to transform lines like:
#   abc\
#   def
# into:
#   abcdef
RX_LINE_CONT = re.compile("[\\\\]\\s*\n\\s*", re.MULTILINE)


def scan(
    provenance: model.SourceLine | None,
    input: str,
    snippet_by_name: model.SnippetByName | None = None,
    debug: bool = False,
) -> model.SourceLines:
    output: model.SourceLines = []
    includes: set[str] = set()

    # stitch continuated lines
    input = RX_LINE_CONT.sub("", input)

    if provenance is None:
        provenance = model.SourceLine("", provenance, None, 0)
    _scan(input, provenance, output, snippet_by_name, includes)

    if debug:
        dprint("=" * 40)
        dprint(provenance)
        dprint("----------")
        dprint(input)
        dprint("----------")
        for l in output:
            dprint(model.repr(l))
        dprint("=" * 40)

    return output


def _scan(
    input: str,
    parent: model.SourceLine,
    output: model.SourceLines,
    snippet_by_name: model.SnippetByName | None,
    includes: set[str],
) -> None:
    for nr, line in enumerate(input.splitlines()):
        if not line.strip():
            continue
        source_line = model.SourceLine(line, line, parent, nr)
        pair = line.split(maxsplit=1)
        if len(pair) == 2 and pair[0] == "#include":
            include(line, source_line, output, snippet_by_name, includes)
        else:
            output.append(source_line)


def include(
    line: str,
    parent: model.SourceLine,
    output: model.SourceLines,
    snippet_by_name: model.SnippetByName | None,
    includes: set[str],
) -> None:
    pair = line.split(maxsplit=1)
    name = pair[1]
    prefix = model.mk_err_prefix_from(parent)

    if name in includes:
        raise model.DfdException(f'{prefix}Recursive include of "{name}"')
    includes.add(name)

    caller = model.SourceLine("", f"<snippet {name}>", parent, 0)
    if name.startswith("#"):
        # include from MD snippet
        if not snippet_by_name:
            raise model.DfdException(
                f"{prefix}source is not markdown, "
                f'cannot include snippet "{name}".'
            )
        name0 = name
        name = name[1:]
        snippet = snippet_by_name.get(name) or snippet_by_name.get(name0)
        if not snippet:
            raise model.DfdException(
                f'{prefix}included snippet "{name}" not found.'
            )

        _scan(snippet.text, caller, output, snippet_by_name, includes)

    else:
        # include from file
        if not os.path.exists(name):
            raise model.DfdException(
                f'{prefix}included file "{name}" not found.'
            )
        with open(name, encoding="utf-8") as f:
            text = f.read()
        _scan(text, caller, output, snippet_by_name, includes)



"""Run the generation process"""

import os.path
import pprint
import re
import textwrap
from typing import Any, Optional

from . import dependency_checker
from . import dfd_dot_templates as TMPL
from . import dot, model, parser, scanner
from .console import dprint


def build(
    provenance: model.SourceLine,
    dfd_src: str,
    output_path: str,
    options: model.Options,
    snippet_by_name: model.SnippetByName | None = None,
) -> None:
    """Take a DFD source and build the final image or document"""
    lines = scanner.scan(provenance, dfd_src, snippet_by_name, options.debug)
    statements, dependencies, attribs = parser.parse(lines, options.debug)
    if dependencies and not options.no_check_dependencies:
        dependency_checker.check(dependencies, snippet_by_name, options)

    items_by_name = parser.check(statements)

    statements = handle_filters(statements, options.debug)
    statements = remove_unused_hidables(statements)
    statements, graph_options = handle_options(statements)

    if options.no_graph_title:
        title = ""
    else:
        title = os.path.splitext(output_path)[0]

    gen = Generator(graph_options, attribs)
    text = generate_dot(gen, title, statements, items_by_name)
    dprint(text)
    dot.generate_image(graph_options, text, output_path, options.format)


def wrap(text: str, cols: int) -> str:
    res: list[str] = []
    for each in text.strip().split("\\n"):
        res += textwrap.wrap(each, width=cols, break_long_words=False) or [""]
    return "\\n".join(res)


class Generator:
    RX_NUMBERED_NAME = re.compile(r"(\d+[.])(.*)")

    def __init__(
        self, graph_options: model.GraphOptions, attribs: model.Attribs
    ) -> None:
        self.lines: list[str] = []
        self.star_nr = 0
        self.frame_nr = 0
        self.graph_options = graph_options
        self.attribs = attribs
        self.attribs_rx = self._compile_attribs_names(attribs)

    def append(self, line: str, statement: model.Statement) -> None:
        self.lines.append("")
        text = model.pack(statement.source.text)
        self.lines.append(f"/* {statement.source.line_nr}: {text} */")
        self.lines.append(line)

    def generate_item(self, item: model.Item) -> None:
        copy = model.Item(**item.__dict__)
        hits = self.RX_NUMBERED_NAME.findall(copy.text)
        if hits:
            copy.text = "\\n".join(hits[0])

        copy.text = wrap(copy.text, self.graph_options.item_text_width)
        attrs = copy.attrs or ""
        attrs = self._expand_attribs(attrs)

        match copy.type:
            case model.PROCESS:
                if self.graph_options.is_context:
                    shape = "circle"
                    fc = "white"
                else:
                    shape = "ellipse"
                    fc = '"#eeeeee"'
                line = (
                    f'"{copy.name}" [shape={shape} label="{copy.text}" '
                    f"fillcolor={fc} style=filled {attrs}]"
                )
            case model.CONTROL:
                fc = '"#eeeeee"'
                line = (
                    f'"{copy.name}" [shape=ellipse label="{copy.text}" '
                    f'fillcolor={fc} style="filled,dashed" {attrs}]'
                )
            case model.ENTITY:
                line = (
                    f'"{copy.name}" [shape=rectangle label="{copy.text}" '
                    f"{attrs}]"
                )
            case model.STORE:
                d = self._attrib_to_dict(copy, attrs)
                line = TMPL.STORE.format(**d)
            case model.NONE:
                line = f'"{copy.name}" [shape=none label="{copy.text}" {attrs}]'
            case model.CHANNEL:
                d = self._attrib_to_dict(copy, attrs)
                if self.graph_options.is_vertical:
                    line = TMPL.CHANNEL_HORIZONTAL.format(**d)
                else:
                    line = TMPL.CHANNEL.format(**d)
            case _:
                prefix = model.mk_err_prefix_from(copy.source)
                raise model.DfdException(
                    f"{prefix}Unsupported item type " f'"{copy.type}"'
                )
        self.append(line, item)

    def _attrib_to_dict(self, item: model.Item, attrs: str) -> dict[str, str]:
        d = self._item_to_html_dict(item)
        d.update({"fontcolor": "black", "color": "black"})
        attrs_d = {
            k: v for k, v in [each.split("=", 1) for each in attrs.split()]
        }
        d.update(attrs_d)
        return d

    def _item_to_html_dict(self, item: model.Item) -> dict[str, Any]:
        d = item.__dict__
        d["text"] = d["text"].replace("\\n", "<br/>")
        return d

    def _compile_attribs_names(
        self, attribs: model.Attribs
    ) -> re.Pattern[str] | None:
        if not attribs:
            return None
        names = ["\\b" + re.escape(k) + "\\b" for k in attribs.keys()]
        pattern = "|".join(names)
        return re.compile(pattern)

    def _expand_attribs(self, attrs: str) -> str:
        def replacer(m: re.Match[str]) -> str:
            alias = m[0]
            if alias not in self.attribs:
                raise model.DfdException(
                    f"Alias "
                    f'"{alias}" '
                    f"not found in "
                    f"{pprint.pformat(self.attribs)}"
                )

            return self.attribs[alias].text

        return (
            self.attribs_rx.sub(replacer, attrs) if self.attribs_rx else attrs
        )

    def generate_star(self, text: str) -> str:
        text = wrap(text, self.graph_options.item_text_width)
        star_name = f"__star_{self.star_nr}__"
        line = f'"{star_name}" [shape=none label="{text}" {TMPL.DOT_FONT_EDGE}]'
        self.lines.append(line)
        self.star_nr += 1
        return star_name

    def generate_connection(
        self,
        conn: model.Connection,
        src_item: model.Item | None,
        dst_item: model.Item | None,
    ) -> None:
        text = conn.text or ""
        text = wrap(text, self.graph_options.connection_text_width)

        src_port = dst_port = ""

        if not src_item:
            src_name = self.generate_star(text)
            text = ""
        else:
            src_name = src_item.name
            if src_item.type == model.CHANNEL:
                src_port = ":x:c"

        if not dst_item:
            dst_name = self.generate_star(text)
            text = ""
        else:
            dst_name = dst_item.name
            if dst_item.type == model.CHANNEL:
                dst_port = ":x:c"

        attrs = f'label="{text}"'

        if conn.attrs:
            attrs += " " + self._expand_attribs(conn.attrs)

        match conn.type:
            case model.FLOW:
                if conn.reversed:
                    attrs += " dir=back"
            case model.BFLOW:
                attrs += " dir=both"
            case model.CFLOW:
                if conn.reversed:
                    attrs += " dir=back"
                    attrs += " arrowtail=normalnormal"
                else:
                    attrs += " arrowhead=normalnormal"
            case model.UFLOW:
                attrs += " dir=none"
            case model.SIGNAL:
                if conn.reversed:
                    attrs += " dir=back"
                attrs += " style=dashed"
            case _:
                prefix = model.mk_err_prefix_from(conn.source)
                raise model.DfdException(
                    f"{prefix}Unsupported connection type " f'"{conn.type}"'
                )
        if conn.relaxed:
            attrs += " constraint=false"

        line = f'"{src_name}"{src_port} -> "{dst_name}"{dst_port} [{attrs}]'
        self.append(line, conn)

    def generate_style(self, style: model.Style) -> None:
        pass

    def generate_frame(self, frame: model.Frame) -> None:
        self.append(f"subgraph cluster_{self.frame_nr} {{", frame)
        self.frame_nr += 1

        self.lines.append(f'  label="{frame.text}"')
        if frame.attrs:
            attrs = self._expand_attribs(frame.attrs)
            self.lines.append(f"  {attrs}")

        for item in frame.items:
            self.lines.append(f'  "{item}"')
        self.lines.append("}")

    def generate_dot_text(self, title: str) -> str:
        graph_params = []

        if self.graph_options.is_context:
            graph_params.append(TMPL.GRAPH_PARAMS_CONTEXT_DIAGRAM)

        if title:
            graph_params.append(TMPL.DOT_GRAPH_TITLE.format(title=title))
        else:
            graph_params.append(TMPL.DOT_GRAPH_NOTITLE)

        if self.graph_options.is_vertical:
            graph_params.append("rankdir=TB")
        else:
            graph_params.append("rankdir=LR")

        if self.graph_options.is_rotated:
            graph_params.append(f"rotate=90")

        block = "\n".join(self.lines).replace("\n", "\n  ")
        text = TMPL.DOT.format(
            title=title,
            block=block,
            graph_params="\n  ".join(graph_params),
        ).replace("\n  \n", "\n\n")
        # dprint(text)
        return text


def generate_dot(
    gen: Generator,
    title: str,
    statements: model.Statements,
    items_by_name: dict[str, model.Item],
) -> str:
    """Iterate over statements and generate a dot source file"""

    def get_item(name: str) -> Optional[model.Item]:
        return None if name == "*" else items_by_name[name]

    for statement in statements:
        match statement:
            case model.Item() as item:
                gen.generate_item(item)

            case model.Connection() as connection:
                src_item = get_item(connection.src)
                dst_item = get_item(connection.dst)
                gen.generate_connection(connection, src_item, dst_item)

            case model.Style() as style:
                gen.generate_style(style)

            case model.Frame() as frame:
                gen.generate_frame(frame)

    return gen.generate_dot_text(title)


def remove_unused_hidables(statements: model.Statements) -> model.Statements:
    # collect used items
    connected_items = set()
    for statement in statements:
        match statement:
            case model.Connection() as conn:
                pass
            case _:
                continue
        for point in conn.src, conn.dst:
            connected_items.add(point)

    # filter out unconnected items
    new_statements = []
    for statement in statements:
        match statement:
            case model.Item() as item:
                if item.hidable and item.name not in connected_items:
                    continue
        new_statements.append(statement)

    return new_statements


def handle_options(
    statements: model.Statements,
) -> tuple[model.Statements, model.GraphOptions]:
    new_statements = []
    options = model.GraphOptions()
    for statement in statements:
        prefix = model.mk_err_prefix_from(statement.source)
        match statement:
            case model.Style() as style:
                match style.style:
                    case "vertical":
                        options.is_vertical = True
                    case "context":
                        options.is_context = True
                    case "horizontal":
                        options.is_vertical = False
                    case "rotated":
                        options.is_rotated = True
                    case "unrotated":
                        options.is_rotated = False
                    case "item-text-width":
                        try:
                            options.item_text_width = int(style.value)
                        except ValueError as e:
                            raise model.DfdException(f'{prefix}{e}"')
                    case "connection-text-width":
                        try:
                            options.connection_text_width = int(style.value)
                        except ValueError as e:
                            raise model.DfdException(f'{prefix}{e}"')

                    case _:
                        raise model.DfdException(
                            f"{prefix}Unsupported style " f'"{style.style}"'
                        )

                continue
        new_statements.append(statement)

    return new_statements, options


def find_neighbors(
    filter: model.Filter,
    statements: model.Statements,
    max_neighbors: int,
    debug: bool,
) -> tuple[set[str], set[str]]:
    """Collect names from filter statements of the given type."""

    def _find_neighbors(
        names: set[str], find_down: bool, nreverse: bool
    ) -> set[str]:
        found_names: set[str] = set()
        for statement in statements:
            match statement:
                case model.Connection() as conn:
                    src, dst = conn.src, conn.dst
                    if conn.reversed and not nreverse:
                        src, dst = dst, src

                    if conn.type in (model.BFLOW, model.UFLOW):
                        if dst in names:
                            found_names.add(src)
                        if src in names:
                            found_names.add(dst)
                    else:
                        if find_down:
                            if src in names:
                                found_names.add(dst)
                        else:
                            if dst in names:
                                found_names.add(src)
                case _:
                    continue
        return found_names

    def _nb(nb: int) -> int:
        if nb < 0:
            return max_neighbors
        return nb

    # find down and up neighbors by successive waves of connections
    def _find_neighbors_in_dir(
        fn: model.FilterNeighbors, down: bool
    ) -> set[str]:
        names = set(filter.names)
        neighbor_names: set[str] = set()
        for i in range(_nb(fn.distance)):
            names = _find_neighbors(
                names, find_down=down, nreverse=fn.layout_dir
            )
            if not names:
                break
            dprint(f"  - {i} {down} {fn}")
            dprint(f"     :", neighbor_names)
            dprint(f"   + :", names)
            neighbor_names.update(names)
            dprint(f"   = :", neighbor_names)
        return neighbor_names

    return _find_neighbors_in_dir(
        filter.neighbors_down, down=True
    ), _find_neighbors_in_dir(filter.neighbors_up, down=False)


def handle_filters(
    statements: model.Statements, debug: bool = False
) -> model.Statements:
    all_names = set([s.name for s in statements if isinstance(s, model.Item)])
    kept_names: set[str] | None = None

    def _check_names(names: set[str], in_names: set[str], prefix: str) -> None:
        if not names.issubset(all_names):
            diff = ", ".join(names - all_names)
            raise model.DfdException(f'{prefix} Name(s) unknown: {diff}')

        if not names.issubset(in_names):
            diff = ", ".join(names - in_names)
            raise model.DfdException(
                f'{prefix} Name(s) no longer available due to previous filters: {diff}'
            )

    replacement: dict[str, str] = {}
    skip_frames_for_names: set[str] = (
        set()
    )  # names for which frames should be skipped due to neighbors_only

    # collect filtered names
    for statement in statements:
        prefix = model.mk_err_prefix_from(statement.source)

        if isinstance(statement, model.Filter):
            dprint("*** Filter:", statement)
            dprint("    before:", kept_names)

        match statement:
            case model.Only() as f:
                if kept_names is None:
                    kept_names = set()

                # all Only names must be valid
                names = set(f.names)
                _check_names(names, all_names, prefix)

                # add names from this Only statement
                if (
                    not f.neighbors_up.no_anchors
                    and not f.neighbors_down.no_anchors
                ):
                    dprint("ONLY: adding nodes:", names)
                    kept_names.update(f.names)

                # add neighbors
                downs, ups = find_neighbors(
                    f, statements, len(all_names), debug
                )
                dprint("ONLY: adding neighbors:", downs, ups)
                kept_names.update(downs)
                kept_names.update(ups)

                if f.neighbors_up.no_frames:
                    skip_frames_for_names.update(ups)
                    if not f.neighbors_up.no_anchors:
                        skip_frames_for_names.update(names)
                if f.neighbors_down.no_frames:
                    skip_frames_for_names.update(downs)
                    if not f.neighbors_down.no_anchors:
                        skip_frames_for_names.update(names)

            case model.Without() as f:
                if kept_names is None:
                    kept_names = all_names.copy()

                # all Without names must be valid
                names = set(f.names)
                names_to_check = names.copy()
                if f.replaced_by:
                    names_to_check.add(f.replaced_by)
                    for name in names:
                        replacement[name] = f.replaced_by
                _check_names(names_to_check, kept_names, prefix)

                # remove names from this Without statement
                if (
                    not f.neighbors_up.no_anchors
                    and not f.neighbors_down.no_anchors
                ):
                    dprint("WITHOUT: removing nodes:", names)
                    kept_names.difference_update(names)

                # remove neighbors
                downs, ups = find_neighbors(
                    f, statements, len(all_names), debug
                )
                dprint("WITHOUT: removing neighbors:", downs, ups)
                kept_names.difference_update(downs)
                kept_names.difference_update(ups)

                if f.neighbors_up.no_frames:
                    skip_frames_for_names.update(ups)
                    if not f.neighbors_up.no_anchors:
                        skip_frames_for_names.update(names)
                if f.neighbors_down.no_frames:
                    skip_frames_for_names.update(downs)
                    if not f.neighbors_down.no_anchors:
                        skip_frames_for_names.update(names)

        if isinstance(statement, model.Filter):
            dprint("    after:", kept_names)

    kept_names = kept_names if kept_names is not None else all_names

    dprint("\nItems to keep", kept_names)

    # apply filters
    new_statements: list[model.Statement] = []
    replaced_connections: dict[str, model.Connection] = {}
    for statement in statements:
        dprint(f"\nHandling statement: {statement}")
        match statement:
            case model.Item() as item:
                # skip nodes that are not in the only list
                if item.name not in kept_names:
                    dprint("=> Skipping item: its name is not in the kept list")
                    continue

            case model.Connection() as conn:
                if conn.src in replacement or conn.dst in replacement:
                    if conn.src in replacement and conn.dst in replacement:
                        continue  # skip connections if both ends are replaced by the same name
                    # replace any end
                    if conn.src in replacement:
                        conn.src = replacement[conn.src]
                    if conn.dst in replacement:
                        conn.dst = replacement[conn.dst]
                    replaced_connections[conn.signature()] = conn
                else:
                    # skip connections if either src or dst is not in the kept list
                    if conn.src not in kept_names or conn.dst not in kept_names:
                        dprint(
                            "=> Skipping connection: some end is not in the kept list"
                        )
                        continue

            case model.Frame() as frame:
                # apply replacements
                for name in frame.items:
                    if name in replacement:
                        frame.items.remove(name)
                        frame.items.append(replacement[name])

                # skip frames if none of the items are in the kept list
                names = set(frame.items)
                if not names.intersection(kept_names):
                    dprint("=> Skipping frame: no items are in the kept list")
                    continue
                else:
                    # adjust frame items by removing those not in the kept list
                    new_items = [n for n in frame.items if n in kept_names]
                    dprint(
                        f"=> Adjusting frame items: {frame.items} -> {new_items}"
                    )
                    frame.items = new_items

                    # skip frames if they contain items for which frames should be skipped
                    if set(new_items).intersection(skip_frames_for_names):
                        dprint(
                            "=> Skipping frame: some items are in the skip-frames list"
                        )
                        continue

        # keep statement
        dprint("=> Keeping statement")
        new_statements.append(statement)

    # remove duplicate connections due to replacements
    kept_new_statements = []
    skiped_statements = set()
    for statement in new_statements:
        match statement:
            case model.Connection() as conn:
                sig = conn.signature()
                if sig in skiped_statements:
                    continue
                if sig in replaced_connections:
                    skiped_statements.add(sig)
        kept_new_statements.append(statement)

    return kept_new_statements



"""Graphviz dot-related generation process"""

import subprocess
import sys

from . import model
from .console import print_error


def generate_image(
    graph_options: model.GraphOptions, text: str, output_path: str, format: str
) -> None:
    if graph_options.is_context:
        engine = "neato"  # circo is not as good
    else:
        engine = "dot"

    cmd = [engine, f"-T{format}", f"-o{output_path}"]
    try:
        subprocess.run(cmd, input=text, encoding="utf-8", check=True)
    except subprocess.CalledProcessError as e:
        for n, line in enumerate(text.splitlines()):
            print(f"{n+1:2}: {line}", file=sys.stderr)
        print_error(f"ERROR: {e}")
        sys.exit(1)
    # print('Generated:', output_path, file=sys.stderr)


def check_installed() -> None:
    cmd = ["dot", "-V"]
    try:
        subprocess.run(cmd, stderr=subprocess.DEVNULL)
    except FileNotFoundError as e:
        print(f'ERROR: "Graphviz" seems not installed:', e, file=sys.stderr)
        sys.exit(2)
